#ifndef _Red_black_tree
#define _Red_black_tree

#include <cstddef> // 'NULL'
#include <list>

using namespace std;

template <typename T> class Red_black_tree {
private:
	list<Red_black_tree<T>*> garbage_collector;
	T _content;
	Red_black_tree<T>* _left, * _right, * _super;

	enum { red, black } _color;
	void balance(Red_black_tree<T>*, const Red_black_tree<T>* const);
	void insert(const T&, const Red_black_tree<T>* const);
	void left_rotate();
	void right_rotate();
public:
	Red_black_tree(const T&);
	~Red_black_tree();
	void insert(const T&);
};

template <typename T> void Red_black_tree<T>::balance(Red_black_tree<T>* rbt, const Red_black_tree<T>* const root) {
	if (rbt->_super == root) {
		rbt->_color = black;
		return;
	}
	while (rbt->_super->_color == red) {
		if (rbt->_super == rbt->_super->_super->_left) {
			Red_black_tree* y = rbt->_super->_super->_right;
			if (y && y->_color == red) {
				rbt->_super->_color = black;
				y->_color = black;
				rbt->_super->_super->_color = red;
				rbt = rbt->_super->_super;
			}
			else {
				if (rbt == rbt->_super->_right) {
					rbt = rbt->_super;
					rbt->left_rotate();
				}
				rbt->_super->_color = black;
				rbt->_super->_super->_color = red;
				rbt->_super->_super->right_rotate();
			}
		}
		else {
			Red_black_tree* y = rbt->_super->_super->_left;
			if (y && y->_color == red) {
				rbt->_super->_color = black;
				y->_color = black;
				rbt->_super->_super->_color = red;
				rbt = rbt->_super->_super;
			}
			else {
				if (rbt == rbt->_super->_left) {
					rbt = rbt->_super;
					rbt->right_rotate();
				}
				rbt->_super->_color = black;
				rbt->_super->_super->_color = red;
				rbt->_super->_super->left_rotate();
			}
		}
	}
}

template <typename T> void Red_black_tree<T>::insert(const T& t, const Red_black_tree<T>* const root) {
	if (_content < t) {
		if (!_right) {
			_right = new Red_black_tree<T>(t);
			_right->_super = this;
			garbage_collector.push_back(_right);
			balance(_right, root);
		}
		else _right->insert(t, root);
	}
	else {
		if (!_left) {
			_left = new Red_black_tree<T>(t);
			_left->_super = this;
			garbage_collector.push_back(_left);
			balance(_left, root);
		}
		else _left->insert(t, root);
	}
}

template <typename T> void Red_black_tree<T>::left_rotate() {
	if (!_right) return;
	Red_black_tree<T>* temp = _right;
	if (_right->_left) _right->_left->_super = this;
	_right = _right->_left;
	temp->_left = this;
	Red_black_tree<T>* pseudo_root = this->_super;
	this->_super = temp;
	temp->_super = pseudo_root;
	if (pseudo_root->_left == this) pseudo_root->_left = temp;
	else pseudo_root->_right = temp;
}

template <typename T> void Red_black_tree<T>::right_rotate() {
	if (!_left) return;
	Red_black_tree<T>* temp = _left;
	if (_left->_right) _left->_right->_super = this;
	_left = _left->_right;
	temp->_right = this;
	Red_black_tree<T>* pseudo_root = this->_super;
	this->_super = temp;
	temp->_super = pseudo_root;
	if (pseudo_root->_left == this) pseudo_root->_left = temp;
	else pseudo_root->_right = temp;
}

template <typename T> Red_black_tree<T>::Red_black_tree(const T& t) : _content(t) {
	_left = _right = _super = NULL;
	_color = red;
}

template <typename T> Red_black_tree<T>::~Red_black_tree() {
	typename list<Red_black_tree<T>*>::iterator i;
	for (i = garbage_collector.begin(); i != garbage_collector.end(); i++) delete* i;
}

template <typename T> void Red_black_tree<T>::insert(const T& t) {
	if (_content < t) {
		if (!_right) {
			_right = new Red_black_tree<T>(t);
			_right->_super = this;
			garbage_collector.push_back(_right);
			balance(_right, this);
		}
		else _right->insert(t, this);
	}
	else {
		if (!_left) {
			_left = new Red_black_tree<T>(t);
			_left->_super = this;
			garbage_collector.push_back(_left);
			balance(_left, this);
		}
		else _left->insert(t, this);
	}
	_color = black;
}

#endif
